
/*
 *
 *  Copyright UKQCD Collaboration, October 2000.
 *  Written by Peter Boyle.
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */


extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include "processor.h"
void do_prefetch ( int ptr,int line );
void do_writehint ( int ptr,int line );
void need_cache_line ( int i );
}

#include "registers.h"


void qcdoc_clover_mat_mlt( char *);

int upper = 0;
int human = 0;
int lower = 0;
int prefetch = 0;
int prefetch_output = 0;
int caller_saves = 0;

char name[80]="";
char *underscore=(char *)"";
char procname[80] = "UNSPECIFIED"; /*Default processor type*/

int main ( int argc, char *argv[])
{
  int arg;
  char *c;
  N_Ahead = 1; /*QCDOC .. only fetch one ahead*/

  /*Process command line*/
  while ( ( arg = getopt(argc,argv,"RpouVlL:n:P:")) != EOF){
    switch(arg){
    case 'R': human = 1; break;
    case 'u': upper = 1; break;
    case 'l': lower = 1; break;
    case 'n': if (strlen(optarg)<50) strcpy(name,optarg); break;
    case 'L': underscore=optarg; break;
    case 'p': prefetch = 1; break;
    case 'o': prefetch_output = 1; break;
    case 'P': if (strlen(optarg)<50) strcpy(procname,optarg); break;
    case 'V': caller_saves = 1; break;
    case 'h':
    default: fprintf(stderr,"Usage: %s -[HTCRgsuil] -L<tok> -n routine_name -P proc\t",argv[0]); 
             fprintf(stderr,"\tR -> human readable .S format \n"); 
             fprintf(stderr,"\tu -> upper case name\n"); 
             fprintf(stderr,"\tl -> lower case name\n"); 
	     fprintf(stderr,"\tp -> prefetch\n");
	     fprintf(stderr,"\to -> prefetch output\n");
             fprintf(stderr,"\tL -> append token <tok> to link name of routine\n"); 
	     fprintf(stderr,"\tV -> Break ABI and use all regs Volatile\n");
             exit (1); break;
    }
  }
  strcat(name,underscore);
  if (upper){
    c = name;
    while ( *c != '\0' ){
      *c = toupper(*c);
      c++;
    }
  }
  if (upper){
    c = name;
    while ( *c != '\0' ){
      *c = toupper(*c);
      c++;
    }
  }

  /*Control output according to user set up args*/
  set_processor_optarg(procname);
  if ( caller_saves ) {
    set_abi_caller_save_all();
  }
  set_human_readable(human);
  setup_cmadds();

  /*queue the naive asm code*/
  qcdoc_clover_mat_mlt(name);

  /*Filter through an virtual out of order processor */
  schedule_for_proc();

  /*Dump the resulting code*/
  dump_instruction_queue();

  return(0);
}

  /*
   * Duplicate the QCDSP clover mat mult ... currently not sure whether
   * it makes use of the "stephenson" dirac symmetry. Will look at flop count.
   */

void qcdoc_clover_mat_mlt( char *name)
{

  /****  This section defines all the registers and offsets I need ****/

  /*
   * This marks the argument registers as defined by ABI as off limits
   * to us until they are freed by "getarg()";
   */
  int dum = defargcount(4);

  /*Handle for the loop entry point*/
  int branchno;
  int retno ;

  /* QCDSP code is as follows   */
//--------------------------------------------------------------------------
// calculate Y = A X   where  Y, X:   spinors (checkboarded)
//                               A:   compressed Clover matrices (chbd'ed)
//                               n:   number of local sites (chbd'ed)
//--------------------------------------------------------------------------
// Algorithm:
//   A = A0      A1, -A2     A4, -A5    A9, -A10     A16,-A17    A25, -A26
//       1,2     3           6,-7       11,-12       18,-19      27,-28
//       4,5     6,7         8          13,-14       20,-21      29,-30
//       9,10    11,12       13,14      15           22,-23      31,-32
//       16,17   18,19       20,21      22,23        24          33,-34
//       25,26   27,28       29,30      31,32        33,34       35
//--------------------------------------------------------------------------
// PAB: This matrix is hermitian, and (I guess) best used in the
//      ordering 
//  for (int site = 0; site < n*2; ++site, A+= 36, Y+= 12, X+=12) {
//    k=0;
//    Are = A[k++];
//    Y[0] = Are * X[0];
//      Y[1] = Are * X[1];
//
//      for(i=0;i<6;i++){
//         for(j=0;j<i;j++){
//
//  	 //    Y[j] += Aji Xi
//  	 Are = A[k++];
//  	 Aim = -A[k++];
//  	 Xre = X[i*2];
//  	 Xim = X[i*2+1];
//  	 Y[j*2]   += Are * Xre - Aim * Xim;
//  	 Y[j*2+1] += Aim * Xre + Are * Xim;
//
//  	 Aim = -Aim; // Aij = Aji^*
//  	 Xre = X[j*2];
//  	 Xim = X[j*2+1];
//  	 if ( j == 0 ) {
//  	 //         Y[i] = Aij Xj
//  	   Y[i*2]   = Are * Xre - Aim * Xim;
//  	   Y[i*2+1] = Aim * Xre + Are * Xim;
//  	 } else {
//  	 //         Y[i] += Aij Xj
//  	   Y[i*2]   += Are * Xre - Aim * Xim;
//  	   Y[i*2+1] += Aim * Xre + Are * Xim;
//  	 }
//
//         }
//          // Y[n] += A_nn X[n] 
//         if ( i != 0 ){
//  	 Are = A[k++];
//  	 Y[i*2]  += Are * X[i*2];
//  	 Y[i*2+1]+= Are * X[i*2+1];
//         }
//
//      }
//    }
//  }
//      
// This would then be linear in the (big) A accesses.
// Unfortunately register usage for this is poor. Devotes approx 24
// To in an output spinors leaving 8 for A's. Since we need 3 concurrent
// cmadds to pipeline properly, this means 6 registers will cover 6 cmadds.
// This will probably work on the 440, with fast L1... wont cover L2 well
// on other architectures unfortunately...
// NB This has zero redundant load stores, which since the 440 is L/S starved
// is the reason I went for a less portable implementation. (Will work
// on power,( in principle IA64, mips too ).
// Will try to get around to a sun/alpha implemtation sometime.

//      Y[0] = A[0]*X[0] + A[1]*X[2] + A[2]*X[3] + A[4]*X[4] + A[5]*X[5] 
//        + A[9]*X[6] + A[10]*X[7] + A[16]*X[8] + A[17]*X[9]
//        + A[25]*X[10] + A[26]*X[11]; 
//      Y[1] = A[0]*X[1] + A[1]*X[3] - A[2]*X[2] + A[4]*X[5] - A[5]*X[4] 
//        + A[9]*X[7] - A[10]*X[6] + A[16]*X[9] - A[17]*X[8] 
//        + A[25]*X[11] - A[26]*X[10];
//      Y[2] = A[1]*X[0] - A[2]*X[1] + A[3]*X[2] + A[6]*X[4] + A[7]*X[5] 
//        + A[11]*X[6] + A[12]*X[7] + A[18]*X[8] + A[19]*X[9] 
//        + A[27]*X[10] + A[28]*X[11];
//      Y[3] = A[1]*X[1] + A[2]*X[0] + A[3]*X[3] + A[6]*X[5] - A[7]*X[4] 
//        + A[11]*X[7] - A[12]*X[6] + A[18]*X[9] - A[19]*X[8] 
//        + A[27]*X[11] - A[28]*X[10];
//      Y[4] = A[4]*X[0] - A[5]*X[1] + A[6]*X[2] - A[7]*X[3] 
//        + A[8]*X[4] + A[13]*X[6] + A[14]*X[7] + A[20]*X[8] 
//        + A[21]*X[9] + A[29]*X[10] + A[30]*X[11];
//      Y[5] = A[4]*X[1] + A[5]*X[0] + A[6]*X[3] + A[7]*X[2] 
//        + A[8]*X[5] + A[13]*X[7] - A[14]*X[6] + A[20]*X[9] 
//        - A[21]*X[8] + A[29]*X[11] - A[30]*X[10];
//      Y[6] = A[9]*X[0] - A[10]*X[1] + A[11]*X[2] - A[12]*X[3] 
//        + A[13]*X[4] - A[14]*X[5] + A[15]*X[6] + A[22]*X[8] 
//        + A[23]*X[9] + A[31]*X[10] + A[32]*X[11];
//      Y[7] = A[9]*X[1] + A[10]*X[0] + A[11]*X[3] + A[12]*X[2] 
//        + A[13]*X[5] + A[14]*X[4] + A[15]*X[7] + A[22]*X[9] 
//        - A[23]*X[8] + A[31]*X[11] - A[32]*X[10];
//      Y[8] = A[16]*X[0] - A[17]*X[1] + A[18]*X[2] - A[19]*X[3] 
//        + A[20]*X[4] - A[21]*X[5] + A[22]*X[6] - A[23]*X[7] 
//        + A[24]*X[8] + A[33]*X[10] + A[34]*X[11];
//      Y[9] = A[16]*X[1] + A[17]*X[0] + A[18]*X[3] + A[19]*X[2] 
//        + A[20]*X[5] + A[21]*X[4] + A[22]*X[7] + A[23]*X[6] 
//        + A[24]*X[9] + A[33]*X[11] - A[34]*X[10];
//      Y[10] = A[25]*X[0] - A[26]*X[1] + A[27]*X[2] - A[28]*X[3] 
//        + A[29]*X[4] - A[30]*X[5] + A[31]*X[6] - A[32]*X[7] 
//        + A[33]*X[8] - A[34]*X[9] + A[35]*X[10];
//      Y[11] = A[25]*X[1] + A[26]*X[0] + A[27]*X[3] + A[28]*X[2] 
//        + A[29]*X[5] + A[30]*X[4] + A[31]*X[7] + A[32]*X[6] 
//        + A[33]*X[9] + A[34]*X[8] + A[35]*X[11];


  struct rotating_reg *Aregs;


  reg_array_1d(Y, Fregs, 12 );
  reg_array_1d(X, Fregs, 12 );
  offset_1d(CHI_IMM    ,  12 );
  offset_1d(CLOV_IMM    , 36 );
  int Y_used[12] = { 0,0,0,0,  0,0,0,0,  0,0,0,0};

  /*Needed to read the number of sites to do*/
  alreg(counter,Iregs);
  /*
   * Also need Integer base registers for arrays
   */
  alreg(Abase, Iregs) ;
  alreg(Chiin, Iregs) ;
  alreg(Chiout,Iregs) ;
  alreg(Coprv,Iregs) ;

  /*
   * Define immediate offsets for iterating the U and Chi pointers
   */
  def_off( ZERO_IMM, 0);
  def_off( CHI, 12);
  def_off( MAT, 36);


  int i,j,k,t;
  int co,sp,rei;
  int Are,Aim;

  /*
   * Streams stuff
   */
  struct stream *PreChi;
  struct stream *PreMat;
  struct stream *WH;

                 /*
		  *
		  * Start of the PROLOGUE
		  *
		  */


  make_inst(DIRECTIVE,Enter_Routine,name);


  /*Allocate stack save any callee save registers we need etc...*/
  grab_stack(0);
  save_regs();

  /*Define our arguments - all pointers ala fortran*/
  getarg(Chiout);
  getarg(Abase);
  getarg(Chiin);
  getarg(counter);
  queue_iadd(counter,counter,counter);

  /*Branch to stack restore if length <1*/
  retno = get_target_label();
  check_iterations(counter,retno); 

  /*Define linear access streams*/
  PreChi = create_stream(CHI,Chiin,counter,STREAM_IN,LINEAR);
  PreMat = create_stream(MAT,Abase,counter,STREAM_IN,LINEAR);
  WH     = create_stream(CHI,Chiout,counter,STREAM_OUT,LINEAR);


  Aregs = create_rotating_reg(Fregs,6,"Clov");

          /*
	   *
	   *   Start the MAIN LOOP
	   *
	   */  


#define SPLIT_X 7
#define SPLIT_A 4
#define SPLIT_Y 8

  for ( i = 0; i<SPLIT_X; i++ ) {
    queue_fload(X[i],CHI_IMM[i],Chiin);
  }

  reset_rotating_register(Aregs);
  for ( i = 0; i<SPLIT_A; i++ ) {
    Are = get_rotating_register(Aregs);
    queue_fload(Are,CLOV_IMM[i],Abase);
  }
  queue_iadd_imm(Coprv,Chiout,ZERO_IMM);

  //  pragma(STORE_LIM,1);
  pragma(DCBT_SPACE,6);
  pragma(DCBT_POST,1);
  branchno = start_loop(counter);

  reset_rotating_register(Aregs);

  for(i=0;i<4;i++){

    if ( 2*i >= SPLIT_Y ) 
      queue_fstore(Y[2*i],CHI_IMM[2*i],Coprv);
    if ( 2*i+1 >= SPLIT_Y ) 
      queue_fstore(Y[2*i+1],CHI_IMM[2*i+1],Coprv);


    //Load the X which hasn't been loaded at this value of i
    if ( i*2 >= SPLIT_X ) {
      queue_fload(X[2*i],CHI_IMM[2*i],Chiin);
    }
    if ( i*2+1 >= SPLIT_X ) {
      queue_fload(X[2*i+1],CHI_IMM[2*i+1],Chiin);
    }

  /*
   * ( - * * * - - )
   * ( * - * * - - )
   * ( * * - * - - )
   * ( * * * - - - )
   * ( - - - - - - )
   * ( - - - - - - )
   */
    for(j=0;j<i;j++){
 
      // Load Aij
      // .. N.B. Aji = Aij*
      k = i * (i+2) - (i-j)*2 ;
      Are = get_rotating_register(Aregs);
      Aim = get_rotating_register(Aregs);
      if ( k   >= SPLIT_A) queue_fload(Are,CLOV_IMM[k],Abase);
      if ( k+1 >= SPLIT_A) queue_fload(Aim,CLOV_IMM[k+1],Abase);
      // Yi += Aij Xj
      if ( j == 0 ) {
        queue_cmul (Y[2*i],Y[2*i+1],Are,Aim,X[2*j],X[2*j+1]);
      } else {
        queue_cmadd (Y[2*i],Y[2*i+1],Are,Aim,X[2*j],X[2*j+1]);
      }
      // Yj += Aij* Xi
      queue_conjmadd (Y[2*j],Y[2*j+1],Are,Aim,X[2*i],X[2*i+1]);

    }

  /*
   * ( * - - - - - )
   * ( - * - - - - )
   * ( - - * - - - )
   * ( - - - * - - )
   * ( - - - - - - )
   * ( - - - - - - )
   */
    //Assign the Y which hasn't been started at this value of i
    /* Y[i] = Aii X[i] */
    k = i * (i+2) ;
    Are = get_rotating_register(Aregs);
    if ( k >= SPLIT_A) queue_fload(Are,CLOV_IMM[k],Abase);
    if ( i == 0 ) {
      queue_fmul (Y[2*i],Are,X[2*i]);
      queue_fmul (Y[2*i+1],Are,X[2*i+1]);
    }else {
      queue_fmac (Y[2*i],Are,X[2*i]);
      queue_fmac (Y[2*i+1],Are,X[2*i+1]);
    }

  }


  for (i=4;i<6;i++){
    if ( 2*i >= SPLIT_Y ) 
      queue_fstore(Y[2*i],CHI_IMM[2*i],Coprv);
    if ( 2*i+1 >= SPLIT_Y ) 
      queue_fstore(Y[2*i+1],CHI_IMM[2*i+1],Coprv);
  }
    //Load the X which hasn't been loaded at this value of i
  for ( i=4; i < 6; i++ ) {
    if ( i*2 >= SPLIT_X ) {
      queue_fload(X[2*i],CHI_IMM[2*i],Chiin);
    }
    if ( i*2+1 >= SPLIT_X ) {
      queue_fload(X[2*i+1],CHI_IMM[2*i+1],Chiin);
    }
  }
 

  /*
   * ( - - - - - - )
   * ( - - - - - - )
   * ( - - - - - - )
   * ( - - - - - - )
   * ( - - - - * - )
   * ( - - - - - * )
   */
  for ( i =4 ; i <6; i++) {
    //Assign the Y which hasn't been started at this value of i
    /* Y[i] = Aii X[i] */
    k = i * (i+2) ;
    Are = get_rotating_register(Aregs);
    if ( k >= SPLIT_A) queue_fload(Are,CLOV_IMM[k],Abase);
    queue_fmul (Y[2*i],Are,X[2*i]);
    queue_fmul (Y[2*i+1],Are,X[2*i+1]);
  }

  /*Loop in a funny way: in order to tolerate 5 cycles latency
   *for 440 we must be collecting into multiple result registers 
   *concurrently. Do this by performing 4 cmadds concurrently
   *by looping in the two bands:
   *
   * ( - - - - * * )
   * ( - - - - * * )
   * ( - - - - * * )
   * ( - - - - * * )
   * ( * * * * - * )
   * ( * * * * * - )
   *
   * Only need to do this now since the scheduler picks up the 
   * slop pretty well in the earlier code, but end up with stalls
   * at the exit of the loop.
   */

  for(j=0;j<6;j++){
    for ( i=4;i<6;i++){
      if ( j < i ){
      // Load Aij
      // .. N.B. Aji = Aij*
      k = i * (i+2) - (i-j)*2 ;
      Are = get_rotating_register(Aregs);
      Aim = get_rotating_register(Aregs);
      if ( k   >= SPLIT_A) queue_fload(Are,CLOV_IMM[k],Abase);
      if ( k+1 >= SPLIT_A) queue_fload(Aim,CLOV_IMM[k+1],Abase);
      // Yi += Aij Xj
      queue_cmadd (Y[2*i],Y[2*i+1],Are,Aim,X[2*j],X[2*j+1]);
      // Yj += Aij* Xi
      queue_conjmadd (Y[2*j],Y[2*j+1],Are,Aim,X[2*i],X[2*i+1]);
      }
    }
  }

  queue_iadd_imm(Coprv,Chiout,ZERO_IMM);

  for ( i = 0 ; i < SPLIT_Y ; i++ ) {
    queue_fstore(Y[i],CHI_IMM[i],Coprv);
  }

  if ( prefetch ) {
    queue_prefetch(PreChi); 
    queue_prefetch(PreMat); 
  }
  if ( prefetch_output ) {
    queue_write_hint(WH); 
  }

  iterate_stream(PreChi);
  iterate_stream(PreMat);
  iterate_stream(WH);

  for ( i = 0; i<SPLIT_X; i++ ) {
    queue_fload(X[i],CHI_IMM[i],Chiin);
  }

  reset_rotating_register(Aregs);
  for ( i = 0; i<SPLIT_A; i++ ) {
    Are = get_rotating_register(Aregs);
    queue_fload(Are,CLOV_IMM[i],Abase);
  }

  /* TERMINATION point of the loop*/
  stop_loop(branchno,counter);

  for ( j = SPLIT_Y; j<12; j++ ) {
    queue_fstore(Y[j],CHI_IMM[j],Coprv);
  }


  make_inst(DIRECTIVE,Target,retno);

                  /*
		   *
		   * EPILOGUE
		   *
		   */

  restore_regs();
  free_stack();
  make_inst(DIRECTIVE,Exit_Routine,name);

  return;
}








