/*
 *
 *  Copyright Peter Boyle and Glasgow University, 2000.
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include "processor.h"



struct processor * PROC = NULL;

/*These routines should be visible internally to this file only*/
void test_proc(void );
void initialise_streams (void );


void set_processor(struct processor *proc)
{
  PROC = proc;
}
void set_abi_caller_save_all(void)
{
  PROC->Isave = 0L;
  PROC->Fsave = 0L;
}
void test_proc(void )
{
  if ( PROC == NULL ) {
    puts("Error: Processor not set");
    exit(0);
  }
}


/* ...}*/

/* {... LOOP management routines*/
int nxtlab = 0;

int get_target_label(void)
{
  return(nxtlab++);
}
/*  
 *Start loop and end loop are in architecture files since
 * inner loop on power pc uses the CTR register for better
 * branch prediction
 */
/*...}*/

/* {...Insert an instruction*/
void make_inst(int PIPE,int OP,...)
{
  va_list ap;
  int i;
  int ivar;
  int ivar2;
  int argtype;
  struct instruction * inst = malloc(sizeof(struct instruction));

  int ignore_alloc;
  if ( !inst ){ perror("Malloc failed"); exit(0);}
  test_proc();

  ignore_alloc = 0;
  if (
        (PIPE == STORPIPE && OP ==SAVE_REG_EA )
     || (PIPE == LOADPIPE && OP ==LOAD_REG_EA ) 
     || (PIPE == FSTRPIPE && OP ==FSAVE_REG_EA)
     || (PIPE == FLODPIPE && OP ==FLOAD_REG_EA) 
       ) ignore_alloc = 1;

  *inst = PROC->iset->instructions[PIPE][OP];



  va_start(ap,OP);

  for (i=0; i<inst->narg;i++){
    argtype = inst->argtypes[i];
    switch( argtype ){

    case Fregs:
      ivar = va_arg(ap,int);
      if ( ivar<0 || ivar > PROC->fregs->regfile ) {
	puts("Error: Invalid floating register");exit(0);break;
      }
      if ( ! PROC->fregs->allocated[ivar] && ! ignore_alloc) {
        puts("Error: unallocated freg used");
        exit(0);
      }
      inst->args[i] = ivar;
      break;
    case Iregs:
      ivar = va_arg(ap,int);
      if ( ivar<0 || ivar > PROC->iregs->regfile ) {
	printf("Error: Invalid register%d\n",ivar);exit(0);break;
      }
      if ( ! PROC->iregs->allocated[ivar] && ! ignore_alloc) {
        printf("Error: unallocated reg %d used",ivar);
        exit(0);
      }
      inst->args[i] = ivar;
      break;
    case SIMM:
      /*Get one number argument*/
      ivar = va_arg(ap,int);
      
      if ( abs(get_offset(ivar)) > 16000 ) {
         puts("Error: Immediate arg");exit(0);break;
      }
      inst->args[i] = ivar ;
      break;
    case Cycles:
      /*Get one number argument*/
      ivar = va_arg(ap,int);
      inst->args[i] = ivar ;
      break;
    case Label:
      /*Get one number argument*/
      if ( inst->pipe == DIRECTIVE ){
        if ( inst->code == Target || inst->code == CmovTarget){
          ivar = va_arg(ap,int);
          if ( ivar >= nxtlab || ivar < 0 ) {
            puts("Error: Unknown Label");exit(0);break;
          }
          inst->args[i] = ivar ;
        }
        else {inst->mnemonic=va_arg(ap,char *);}
      }
      else{
        ivar = va_arg(ap,int);
        if ( ivar >= nxtlab || ivar < 0 ) {
           printf("Error: Unknown label %d,%d:%d %d\n",PIPE,OP,ivar,nxtlab);exit(0);break;
        }
        inst->args[i] = ivar ;
      }
      break;
           
    default:
      puts("Error: bad argtype");exit(0);break;
    }
  }
  va_end(ap);

  /*And add the instruction to the queue*/
  queue_instruction(inst);
  return;
}


/* ...End instruction insertion}*/



int get_I_size(void)
{
  return ( PROC->I_size);
}

/* {...QUEUEING convenience routines */
void queue_load_addr( int dest, int immhandle, int src )
{
  if ( PROC->iset->instructions[LOADPIPE][LOAD_ADDR].mnemonic == NULL ){
    make_inst ( IALUPIPE,IADD_IMM,dest,src,immhandle);
  }
  else {
    make_inst ( LOADPIPE,LOAD_ADDR,dest,immhandle,src);
  }
}
void queue_fload ( int reg, int offhandle, int basereg )
{
  make_inst ( FLODPIPE,FLOAD_EA,reg,offhandle,basereg );
}
void queue_fstore ( int reg, int offhandle, int basereg )
{
  make_inst ( FSTRPIPE,FSAVE_EA,reg,offhandle,basereg );
}
void queue_iload ( int reg, int offhandle, int basereg )
{
  make_inst ( LOADPIPE,ILD_EA,reg,offhandle,basereg );
}
void queue_istore ( int reg, int offhandle, int basereg )
{
  make_inst ( STORPIPE,IST_EA,reg,offhandle,basereg );
}
void queue_iadd( int dest, int a, int b )
{
  make_inst ( IALUPIPE,IADD,dest,a,b );

}
void queue_imul( int dest, int a, int b )
{
  make_inst ( IALUPIPE,IMUL,dest,a,b );

}
void queue_imul_imm( int dest, int a, int himm )
{
 make_inst ( IALUPIPE,IMUL_IMM,dest,a,himm); 

}
void queue_iadd_imm( int dest, int src, int immhandle )
{
  make_inst ( IALUPIPE,IADD_IMM,dest,src,immhandle);
}

void queue_fmul(int dest,int a, int b)
{
  make_inst( FMULPIPE,FMUL,dest,a,b);
}
void queue_fadd(int dest,int a, int b)
{
  make_inst( FADDPIPE,FADD,dest,a,b);
}
void queue_fsub(int dest,int a, int b)
{
  make_inst( FADDPIPE,FSUB,dest,a,b);
}
void queue_fmov(int dest,int src)
{
  make_inst( FMOVPIPE,FMOV,dest,src);
}
void queue_fneg(int dest,int src)
{
  make_inst( FMOVPIPE,FNEG,dest,src);
}
void queue_iload_imm(int dest,int imm)
{
  make_inst( LOADPIPE,LOAD_IMM,dest,imm);
}


/* ...END queueing convenience routines */







void queue_barrier(void)
{
  make_inst(DIRECTIVE,Pipe_Flush, "Barrier");
}
void queue_mbar(void)
{
  make_inst(DIRECTIVE,LS_BARRIER, "mbar");
}

void pragma(int code,int val)
{
  make_inst(PRAGMA,code,val);
}

/* ...}*/




