/*
 *
 *  Copyright Peter Boyle and Glasgow University, 2000.
 *  This software is provided for NON-COMMERCIAL use only,
 *  and may not be redistributed without permission.
 *  It is provided as is and is not guaranteed fit for any purpose.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include "processor.h"

extern struct processor *PROC;
extern struct inst_queue_entry *HEAD;

#define REG_FREED 3

/* { ... REGISTER management routines */
int allocate_reg( int kind,const char *name)
{
  struct regs *set;
  int i;
  test_proc();
  switch ( kind ){
    case Fregs: set = PROC->fregs;break;
    case Iregs: set = PROC->iregs;break;
    default: puts("Bad Arugment: Allocate reg"); exit(0);
  }

  for ( i =0 ; i < set->regfile; i++){
    if ( set->allocated[i]==0 ){
      set->mnemonic[i] = name;
      set->allocated[i] = 1;
      return(i);
    }
  }  

  for ( i =0 ; i < set->regfile; i++){
    if ( set->allocated[i] == REG_FREED ){
      fprintf(stderr,"Warning reusing register %s\n",make_reg_name(kind,i));
      set->mnemonic[i] = name;
      set->allocated[i] = 1;
      return(i);
    }
  }  
  puts("Error: Couldn't allocate register");
  puts(name);
  exit(0);
}

void free_reg( int kind,int num)
{
  struct regs *set;
  test_proc();
  switch ( kind ){
    case Fregs: set = PROC->fregs;break;
    case Iregs: set = PROC->iregs;break;
    default: puts("Bad Arugment: Free reg"); exit(0);
  }

  if ( num > set -> regfile || num < 0 ) {
     puts("Bad Arugment: Free reg"); exit(0);
  }

  if ( !set->allocated[num] ){
    puts("Error: freeing unallocated reg");
    printf("Kind = %d, num = %d",kind,num);
  }
  set->allocated[num] = REG_FREED; /*Tag for reuse but keep mnemonic around unless actually reused*/
  return;
}

/* ...}*/


struct rotating_reg *create_rotating_reg(int kind,int len,const char *name)
{
  struct rotating_reg *rp = malloc(sizeof(struct rotating_reg));
  int i;
  char *rnam;
  int namlen;
  rp -> kind = kind;
  rp -> last = 0;
  rp -> num = len;
  rp -> regs = malloc(len * sizeof(int));   
  namlen = strlen(name+4); /*4 since may port to IA64 with 3 digit regs*/
  for ( i = 0; i<len; i++){
    rnam = malloc(namlen);
    sprintf(rnam,"%s%d",name,i);
    rp->regs[i] = allocate_reg(kind,rnam);
  }
  return(rp);
}

int get_rotating_register(struct rotating_reg *rp)
{ 
  int r = rp->regs[rp->last];

  rp->last ++;
  if ( rp->last >= rp->num ){
    rp->last = 0;
  }
  return(r);
}
void reset_rotating_register(struct rotating_reg *rp)
{
  rp->last = 0;
}
 







